//-----------------------------------------------------------------------------
//  DXL_additions.i  -  Support symbols and routines added to the 
//                      Python DXL interface
//                   
//                    Included by DXL.i
//-----------------------------------------------------------------------------

//=============================================================================
// FIXME:
//   1) Need to clean out callback mappings for a connection when 
//      DXLCloseConnection is called; also when a broken connection is detected
//=============================================================================

//-----------------------------------------------------------------------------
//  New convenience constants
//-----------------------------------------------------------------------------

#define DXL_PACKETTYPE_LAST PACK_LINK

%{
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
%}

//-----------------------------------------------------------------------------
//  A DXL Module Exception
//-----------------------------------------------------------------------------

%{
PyObject *DXLExc_Error;            /*  DXL module exception type for Python.  */
                                   /*    Convention says call it <m>[Ee]rror  */
                                   /*    but DXL already has DXLError.        */
%}

%init %{
  /*  New exceptions for the DXL module  */
	DXLExc_Error = PyErr_NewException( "DXL.DXError", NULL, NULL );
	if ( DXLExc_Error != NULL )
		PyDict_SetItemString( d, "DXError", DXLExc_Error );
%}

%inline %{
  void DXError( void ) {}          /*  This is a really cheap trick to get    */
                                   /*    SWIG to add a DXError = DXLc.DXError */
                                   /*    entry in our shadow class for the    */
                                   /*    DXError exception we create in the   */
                                   /*    module init function (see above).    */
%}

//-----------------------------------------------------------------------------
//  WaitForDXIdle( conn ) - new function
//-----------------------------------------------------------------------------

%inline %{
  
void WaitForDXIdle( DXLConnection *conn )
  /*  Wait for the DX server to finish execution.                             */
  /*                                                                          */
  /*    Note that DX may "never" say it's finished if it encounters an        */
  /*    error.  See below for details.   You should detect this and           */
  /*    throw an exception if this occurs.                                    */
  /*                                                                          */
  /*  THROWS:                                                                 */
  /*     DXError      -  If DX dies, select() will fail because the dxexec    */
  /*                     file descriptor is toast.  We throw DXError in       */
  /*                     response (see the typemap for this routine in        */
  /*                     DXL_typemaps.i).                                     */
{
  int       status = 1;
  int       fd = DXLGetSocket( conn );
  DXLError  error;
  fd_set    fds;

  FD_ZERO( &fds );
  FD_SET( fd, &fds );

  /*  The smart way - doesn't burn CPU and the algorithm (though not this  */
  /*   routine specifically) coexists with event-driven apps -- via        */
  /*   XtAddInput(), Tk fileevent, etc.                                    */

  while ( status ) {

    if ( DXLIsMessagePending( conn ) ) {

      PyErr_Clear();

      error = DXLHandlePendingMessages( conn );

      /*  If a Python callback threw an exception, service it in the  */
      /*    except typemap for this routine.                          */
      if ( PyErr_Occurred() )
        return;

      if ( error != OK ) {
        strcpy( my_error_msg, "DXLHandlePendingMessages() failed" );
        my_error = 1;
        return;
      }
    }

    error = DXLGetExecutionStatus( conn, &status );
    if ( error != OK ) {
      strcpy( my_error_msg, "DXLGetExecutionStatus() failed" );
      my_error = 1;
      return;
    }

    /*  Wait on more messages from DX until execution status is idle.  */
    /*    This is blocked most of the time (versus a busy-wait poll),  */
    /*    so it doesn't eat big chunks of CPU.                         */

    /*  IMPORTANT!  NOTE!!!: If a DX error occurs while DX is executing, an   */
    /*    error handler cb and/or ERROR message will be sent to the           */
    /*    application, but the DX execution status may stay "stuck on"        */
    /*    (i.e. DXLGetExecutionStatus() == 1) even after all the messages     */
    /*    have been sent, received, and processed.  DX is done executing and  */
    /*    is in an error state, but we don't know that here.                  */
    /*                                                                        */
    /*    So, when a DX error occurs, we may get stuck in this loop if the    */
    /*    application isn't killing/restarting DX, terminating the app, or    */
    /*    at least throwing exceptions when DX gets into an error state.      */
    /*    DX is effectively idle; we don't know it here though.               */
    /*                                                                        */
    /*    In the case of a syntax error, ERROR MessageHandler messages (but   */
    /*    no ErrorHandler event) is passed to the app before DX goes into     */
    /*    zombie mode.                                                        */
    /*                                                                        */
    /*    In the future, we may want to consider linking error and message    */
    /*    ERROR reception to this loop so we terminate this rtn with an       */
    /*    exception rather than loop forever when DX goes into zombie mode.   */
    /*    The question is whether all error and message ERROR callbacks mean  */
    /*    that DX is hosed and needs to be killed with extreme prejudice or   */
    /*    not.  We don't know, ...                                            */
    /*                                                                        */
    /*    So for now, we leave it to the application to throw an exception    */
    /*    inside the error and message callbacks when a DX error is received. */
    /*    We recognize that exception being thrown in this routine, and bail  */
    /*    out of the wait loop.                                               */

    if ( status ) {
      int            ret;
      struct timeval timeout;
      
      timeout.tv_sec  = 0;
      timeout.tv_usec = 1000000 * 0.5;             /*  0.5 sec  */

      ret = select( FD_SETSIZE, &fds, NULL, NULL, &timeout );
      if ( ret < 0 ) {
        snprintf( my_error_msg, sizeof( my_error_msg )-1,
                  "select() failed: %s", strerror(errno) );
        my_error_msg[ sizeof( my_error_msg )-1 ] = '\0';
        my_error = 1;
        return;
      }
      /* if ( ret == 0 ) printf( "WaitForDXIdle: timeout (nothing to do)\n" );*/
    }
  }
}

%}

//-----------------------------------------------------------------------------
//  WaitForDXIdleAndHandleError()  wrapper-internal utility routine
//-----------------------------------------------------------------------------

%{

static int WaitForDXIdleAndHandleError( DXLConnection *conn )
  /*  A handy little routine that waits for DX to become IDLE, just like    */
  /*   WaitForDXIdle, but also handles the details of detecting and         */
  /*   propagating Python exceptions, should one be thrown by a Python      */
  /*   DX callback (invoked by a DXLHandlePendingMessages dispatch).        */
  /*  NOTE:  You'll notice this logic is similar to the except typemap      */
  /*   for WaitForDXIdle().                                                 */
  /*                                                                        */
  /*                                                                        */
  /*  RETURNS:  0 = exception occurred; 1 = success.                        */
{
  my_error        = 0;
  my_error_msg[0] = '\0';
  WaitForDXIdle( conn );
  if ( my_error ) {
    PyErr_SetString( DXLExc_Error, my_error_msg );
    return 0;
  }

  /*  If a Python callback threw an exception, restart its propagation  */
  /*    in the controlling Python.                                      */
  if ( PyErr_Occurred() )
    return 0;

  return 1;
}

%}
  
//-----------------------------------------------------------------------------
//  DXLSetSynchronization()  a new one that actually works
//-----------------------------------------------------------------------------

// The DXL-provided DXLSetSynchronization is broken.  If something goes foul
//   (DX server error), then the call you made that triggered it will hang!
//   You'll be a useless process just waiting to be killed.  Useful, eh?
//
// So we replace it with an implementation that does something worthwhile.  
//   Basically, this new method sets an wrapper variable that determines
//   whether the typemap except handlers for DXL wait until DX is idle
//   before returning, dispatching any messages that come back in the
//   meantime (e.g.  Error messages).  So if you registered an error callback
//   and synchronization is on, it'll get called before your function
//   which triggers the error returns, so you'll see the error, and get a
//   chance to exit on error if you've added that logic to your callback.
//
// FIXME:  Note that we go cheap here and ignore the connection argument.  
//   That is, this API now enables/disables synchronization for ALL
//   connection, not just one.

%rename DXLSetSynchronization_NEW DXLSetSynchronization;

%inline %{

static int PyDXLSyncMode = 0;

void DXLSetSynchronization_NEW( DXLConnection *conn, const int onoff )
{
  PyDXLSyncMode = onoff;
}
  
%}
  
//-----------------------------------------------------------------------------
//  DXLGetExecutionStatus() a new one that returns status and not DXLError
//-----------------------------------------------------------------------------

%rename DXLGetExecutionStatus_NEW DXLGetExecutionStatus;

%inline %{

int DXLGetExecutionStatus_NEW( DXLConnection *conn )
{
  int      status = 0;
  DXLError ret = DXLGetExecutionStatus( conn, &status );

  if ( ret != OK ) {
    strcpy( my_error_msg, "DXLGetExecutionStatus() failed" );
    my_error = 1;
  }
  return status;
}
  
%}
  
//-----------------------------------------------------------------------------
//  DX Callback Management Routines:
//    
//      DXLSetErrorHandler            ( conn, callable )
//      DXLSetMessageHandler          ( conn, type, matchstr, callable )
//      DXLRemoveMessageHandler       ( conn, type, matchstr )
//      DXLSetValueHandler            ( conn, name, callable )
//      DXLRemoveValueHandler         ( conn, name )
//      DXLSetBrokenConnectionCallback( conn, callable )
//
//  NOTE:
//    In order to plug our own C wrapper functions in place of the
//    SWIG automatically generated wrappers, the DXL C functions are 
//    renamed to other names in DXL_typemaps.i for the purpose of wrapping
//    e.g.:
//
//        %rename DXLSetErrorHandler              DXL_C_SetErrorHandler;
//
//    causing SWIG's wrappers to advertise this modified name for its wrappers.
//    And then here, we rename the the C versions of our wrappers into
//    the original names e.g.:
//
//        %rename DXLSetErrorHandler_NEW          DXLSetErrorHandler;
//
//-----------------------------------------------------------------------------

%rename DXLSetErrorHandler_NEW              DXLSetErrorHandler;
%rename DXLSetMessageHandler_NEW            DXLSetMessageHandler;
%rename DXLRemoveMessageHandler_NEW         DXLRemoveMessageHandler;
%rename DXLSetValueHandler_NEW              DXLSetValueHandler;
%rename DXLRemoveValueHandler_NEW           DXLRemoveValueHandler;
%rename DXLSetBrokenConnectionCallback_NEW  DXLSetBrokenConnectionCallback;


// Callbacks take a little extra effort because we need to have C calling
//   back to Python.  This takes a middle-man to do the hand-off

%{

#include <string.h>

typedef enum { Hdlr_Error, Hdlr_Message, 
               Hdlr_Value, Hdlr_BrokenConnection } PyDXLHdlrType;

typedef struct {
  PyDXLHdlrType       cb_type;
  DXLConnection      *conn;            /*  For ALL handlers      */
  PyObject           *py_callback;     /*  For ALL handlers      */
  DXLPacketTypeEnum   packet_type;     /*  For message handlers  */
  char                match_str[40];   /*  For message handlers  */
  char                label[40];       /*  For value handlers    */
} PyDXLHdlr;

typedef struct {
  PyDXLHdlr **list;
  int         num;                     /*  Number of used elements  */
  int         size;                    /*  Allocated size of list   */
} PyDXLHdlrList;

static PyDXLHdlrList PyDXLHandler;

%}

// Define some C callback routines to be called by DX.  These
//   lookup the appropriate Python callable object to relay to.

%{
  
static int PythonDXLIsValidHandlerRef( PyDXLHdlr *hdlr_def )
{
  int i;

  for ( i = 0; i < PyDXLHandler.num; i++ )
    if ( PyDXLHandler.list[i] == hdlr_def )
      break;

  return ( i < PyDXLHandler.num );
}

static PyDXLHdlr *PythonDXLNewHandler()
  /*  Create a new handler record in the handler list.  */
{
  int         new_size;
  PyDXLHdlr **list = PyDXLHandler.list;
  PyDXLHdlr  *hdlr_def;

  /*  Grow list  */
  if ( PyDXLHandler.num >= PyDXLHandler.size ) {
    new_size = PyDXLHandler.size + 20;
    list = realloc( list, new_size * sizeof( list[0] ) );
    if ( !list ) {
      strcpy( my_error_msg, "Out of memory" );
      return NULL;
    }
    PyDXLHandler.list = list;
    PyDXLHandler.size = new_size;
  }

  /*  New callback record  */
  hdlr_def = calloc( 1, sizeof( PyDXLHdlr ) );
  if ( !hdlr_def ) {
    strcpy( my_error_msg, "Out of memory" );
    return NULL;
  }
  
  PyDXLHandler.list[ PyDXLHandler.num++ ] = hdlr_def;

  return hdlr_def;
}
  
  
static DXLError PythonDXLDeleteHandler( PyDXLHdlr *hdlr_def )
{
  PyDXLHdlr **list = PyDXLHandler.list;
  int         i;

  for ( i = 0; i < PyDXLHandler.num; i++ )
    if ( PyDXLHandler.list[i] == hdlr_def )
      break;

  if ( i >= PyDXLHandler.num ) {
    strcpy( my_error_msg, 
            "PythonDXLDeleteHandler: No such handler" );
    return ERROR;
  }

  Py_DECREF( list[ i ]->py_callback );
  free( list[ i ] );
  memmove( &list[ i ], &list[ i+1 ], 
           ( (PyDXLHandler.num-1) - i ) * sizeof( list[0] ) );

  PyDXLHandler.num--;
  return OK;
}
  
  
static void PythonDXLMessageHandler( DXLConnection *conn,
                                     const char    *msg, 
                                     void          *data )
  /*  C DX Message and Error callback used to relay to Python callbacks.  */
{
  PyDXLHdlr *hdlr_def = (PyDXLHdlr *) data;
  PyObject  *pyfunc, 
            *pyargs,
            *pyresult;
  char       swig_ptr[128];

  /*printf( "MSGERR CALLBACK: %p '%s' %p\n", conn, msg, data );*/

  /*   First, make sure our callback hasn't been revoked  */
  if ( !PythonDXLIsValidHandlerRef( hdlr_def ) ) {
    PyErr_Format( DXLExc_Error, 
                "ERROR:  DX called a msg/err callback that's been removed!\n" );
    return;
  }

  /*  Sanity check  */
  if (( hdlr_def->cb_type != Hdlr_Error   ) &&
      ( hdlr_def->cb_type != Hdlr_Message )) {
    PyErr_Format( DXLExc_Error, 
                  "ERROR: PythonDXLMessageHandler called for non-error, "
                  "non-message callback (%d).\n", hdlr_def->cb_type );
    return;
  }

  /*  Now pass the buck  */
  pyfunc = hdlr_def->py_callback;

  SWIG_MakePtr( swig_ptr, conn, "_DXLConnection_p" );
  pyargs   = Py_BuildValue( "(ss)", swig_ptr, msg );  /*  Build arg list    */
  pyresult = PyEval_CallObject( pyfunc, pyargs );     /*  Call Python       */
  Py_DECREF( pyargs );                                /*  Trash arg list    */
  if ( pyresult )
    Py_DECREF( pyresult );                            /*  Trash any result  */

  /*  If the callback threw an exception, it will be propagated across   */
  /*    the DXLink layer (we check PyErr_Occurred() on the other side).  */
}

static void PythonDXLValueHandler( DXLConnection *conn,
                                   const char    *label,
                                   const char    *value,
                                   void          *data )
  /*  C DX Value callback used to relay to Python callbacks.  */
{
  PyDXLHdlr *hdlr_def = (PyDXLHdlr *) data;
  PyObject  *pyfunc, 
            *pyargs,
            *pyresult;
  char       swig_ptr[128];

  /*printf( "VALUE CALLBACK: %p %s %s %p\n", conn, label, value, data );*/

  /*   First, make sure our callback hasn't been revoked  */
  if ( !PythonDXLIsValidHandlerRef( hdlr_def ) ) {
    PyErr_Format( DXLExc_Error, 
                "ERROR:  DX called a value callback that's been removed!\n" );
    return;
  }

  /*  Sanity check  */
  if ( hdlr_def->cb_type != Hdlr_Value ) {
    PyErr_Format( DXLExc_Error, 
                  "ERROR: PythonDXLValueHandler called for non-value "
                  "callback (%d).\n", hdlr_def->cb_type );
    return;
  }

  /*  Now pass the buck  */
  pyfunc = hdlr_def->py_callback;

  SWIG_MakePtr( swig_ptr, conn, "_DXLConnection_p" );
  pyargs   = Py_BuildValue( "(sss)", swig_ptr, label,
                            value );                   /*  Build arg list    */
  pyresult = PyEval_CallObject( pyfunc, pyargs );      /*  Call Python       */
  Py_DECREF( pyargs );                                 /*  Trash arg list    */
  if ( pyresult )
    Py_DECREF( pyresult );                             /*  Trash any result  */

  /*  If the callback threw an exception, it will be propagated across   */
  /*    the DXLink layer (we check PyErr_Occurred() on the other side).  */
}

void PythonDXLBrokenConnectionHandler( DXLConnection *conn, 
                                       void          *data )
{
  PyDXLHdlr *hdlr_def = (PyDXLHdlr *) data;
  PyObject  *pyfunc, 
            *pyargs,
            *pyresult;
  char       swig_ptr[128];

  /*printf( "BCONN CALLBACK: %p %p\n", conn, data );*/

  /*   First, make sure our callback hasn't been revoked  */
  if ( !PythonDXLIsValidHandlerRef( hdlr_def ) ) {
    PyErr_Format( DXLExc_Error, 
                  "ERROR:  DX called a broken connection callback that's "
                  "been removed!\n" );
    return;
  }

  /*  Sanity check  */
  if ( hdlr_def->cb_type != Hdlr_BrokenConnection ) {
    PyErr_Format( DXLExc_Error, 
                  "ERROR: PythonDXLBrokenConnectionHandler called for "
                  "       non-broken connection callback (%d).\n", 
                  hdlr_def->cb_type );
    return;
  }

  /*  Now pass the buck  */
  pyfunc = hdlr_def->py_callback;

  SWIG_MakePtr( swig_ptr, conn, "_DXLConnection_p" );
  pyargs   = Py_BuildValue( "(s)", swig_ptr );         /*  Build arg list    */
  pyresult = PyEval_CallObject( pyfunc, pyargs );      /*  Call Python       */
  Py_DECREF( pyargs );                                 /*  Trash arg list    */
  if ( pyresult )
    Py_DECREF( pyresult );                             /*  Trash any result  */

  /*  If the callback threw an exception, it will be propagated across   */
  /*    the DXLink layer (we check PyErr_Occurred() on the other side).  */
}


%}

%inline %{

static DXLError DXLSetErrorHandler_NEW( DXLConnection     *conn, 
                                        PyObject          *pyfunc )
  /* Used to register callable objects in Python as DX error handlers.       */
  /*   To pass client data, use bound methods, lambdas, callable objs, etc.  */
  /*   NOTE:  DX only allows registration of one Error handler callback.     */
{
  PyDXLHdlr *hdlr_def;
  int        i;
  DXLError   ret;

  /*printf( "ERR REGISTER: %p %p\n", conn, pyfunc );*/

  /*  Initialize record  */
  hdlr_def = PythonDXLNewHandler();
  if ( !hdlr_def )
    return ERROR;

  hdlr_def->cb_type     = Hdlr_Error;
  hdlr_def->conn        = conn;
  hdlr_def->py_callback = pyfunc;
  Py_INCREF( hdlr_def->py_callback );

  /*  Register with DX   */
  ret = DXLSetErrorHandler( conn, PythonDXLMessageHandler, hdlr_def );
  if ( ret != OK ) {
    PythonDXLDeleteHandler( hdlr_def );
    strcpy( my_error_msg, "DXLSetErrorHandler failed" );
    return ret;
  }

  /*  Is there a prior matching handler?  If so, nuke it.  */
  for ( i = 0; i < PyDXLHandler.num; i++ )
    if (( PyDXLHandler.list[i]          != hdlr_def   ) &&
        ( PyDXLHandler.list[i]->conn    == conn       ) &&
        ( PyDXLHandler.list[i]->cb_type == Hdlr_Error )) {

      ret = PythonDXLDeleteHandler( PyDXLHandler.list[i] );
      if ( ret != OK )
        return ret;
      break;
    }

  return OK;
}


static DXLError DXLSetMessageHandler_NEW( DXLConnection     *conn,
                                          DXLPacketTypeEnum  type,
                                          const char        *matchstr,
                                          PyObject          *pyfunc )
  /* Used to register callable objects in Python as DX message handlers.    */
  /*   NOTE:  DX maintains one callback per packet type-match string pair.  */
  /*   Different functions and/or different data don't matter.              */
{
  PyDXLHdlr *hdlr_def;
  int        i;
  char      *match_str_safe = (char *) matchstr;
  DXLError   ret;

  /*printf( "MSG REGISTER: %p %d %p %p\n", conn, type, matchstr, pyfunc );*/

  if ( !match_str_safe )
    match_str_safe = "";

  /*  Initialize record  */
  hdlr_def = PythonDXLNewHandler();
  if ( !hdlr_def )
    return ERROR;

  hdlr_def->cb_type     = Hdlr_Message;
  hdlr_def->conn        = conn;
  hdlr_def->py_callback = pyfunc;
  hdlr_def->packet_type = type;
  Py_INCREF( hdlr_def->py_callback );

  hdlr_def->packet_type = type;
  hdlr_def->match_str[0] = '\0';
  strncat( hdlr_def->match_str, match_str_safe, 
           sizeof( hdlr_def->match_str )-1 );

  /*  Register with DX   */
  ret = DXLSetMessageHandler( conn, type, matchstr, PythonDXLMessageHandler, 
                              hdlr_def );
  if ( ret != OK ) {
    PythonDXLDeleteHandler( hdlr_def );
    strcpy( my_error_msg, "DXLSetMessageHandler failed" );
    return ret;
  }

  /*  Is there a prior matching handler?  If so, nuke it.  */
  for ( i = 0; i < PyDXLHandler.num; i++ )
    if (( PyDXLHandler.list[i]              != hdlr_def     ) &&
        ( PyDXLHandler.list[i]->conn        == conn         ) &&
        ( PyDXLHandler.list[i]->cb_type     == Hdlr_Message ) &&
        ( PyDXLHandler.list[i]->packet_type == type         ) &&
        ( strcmp( PyDXLHandler.list[i]->match_str, match_str_safe ) == 0 )) {

      ret = PythonDXLDeleteHandler( PyDXLHandler.list[i] );
      if ( ret != OK )
        return ret;
      break;
    }

  return OK;
}

static DXLError DXLRemoveMessageHandler_NEW( DXLConnection     *conn,
                                             DXLPacketTypeEnum  type,
                                             const char        *matchstr )

  /* Used to unregister callable objects in Python as DX message handlers.  */
  /*   NOTE:  DX maintains one callback per packet type-match string pair.  */
  /*   Different functions and/or different data don't matter.              */
{
  PyDXLHdlr *hdlr_def;
  int        i;
  char      *match_str_safe = (char *) matchstr;
  DXLError   ret;

  /*printf( "MSG UNREGISTER: %p %d %p\n", conn, type, matchstr );*/

  if ( !match_str_safe )
    match_str_safe = "";

  for ( i = 0; i < PyDXLHandler.num; i++ )
    if (( PyDXLHandler.list[i]->conn        == conn         ) &&
        ( PyDXLHandler.list[i]->cb_type     == Hdlr_Message ) &&
        ( PyDXLHandler.list[i]->packet_type == type         ) &&
        ( strcmp( PyDXLHandler.list[i]->match_str, match_str_safe ) == 0 ))
      break;

  if ( i >= PyDXLHandler.num ) {
    ret = ERROR;
    strcpy( my_error_msg, "No matching callback registered" );
  }
  else {

    ret = DXLRemoveMessageHandler( conn, type, matchstr, 
                                   PythonDXLMessageHandler );
    if ( ret != OK )
      strcpy( my_error_msg, "DXLRemoveMessageHandler returned error" );
    else
      ret = PythonDXLDeleteHandler( PyDXLHandler.list[i] );
  }
  return ret;
}

static DXLError DXLSetValueHandler_NEW( DXLConnection     *conn,
                                        const char        *label,
                                        PyObject          *pyfunc )
  /* Used to register callable objects in Python as DX value handlers.      */
{
  PyDXLHdlr *hdlr_def;
  int        i;
  DXLError   ret;

  /*printf( "VAL REGISTER: %p %s %p\n", conn, label, pyfunc );*/

  /*  Initialize record  */
  hdlr_def = PythonDXLNewHandler();
  if ( !hdlr_def )
    return ERROR;

  hdlr_def->cb_type     = Hdlr_Value;
  hdlr_def->conn        = conn;
  hdlr_def->py_callback = pyfunc;
  hdlr_def->label[0]    = '\0';
  strncat( hdlr_def->label, label, sizeof( hdlr_def->label )-1 );

  Py_INCREF( hdlr_def->py_callback );

  /*  Register with DX   */
  ret = DXLSetValueHandler( conn, label, PythonDXLValueHandler, hdlr_def );
  if ( ret != OK ) {
    PythonDXLDeleteHandler( hdlr_def );
    strcpy( my_error_msg, "DXLSetValueHandler failed" );
    return ret;
  }

  /*  Is there a prior matching handler?  If so, nuke it.  */
  for ( i = 0; i < PyDXLHandler.num; i++ )
    if (( PyDXLHandler.list[i]              != hdlr_def     ) &&
        ( PyDXLHandler.list[i]->conn        == conn         ) &&
        ( PyDXLHandler.list[i]->cb_type     == Hdlr_Value   ) &&
        ( strcmp( PyDXLHandler.list[i]->label, label        ) == 0 )) {

      ret = PythonDXLDeleteHandler( PyDXLHandler.list[i] );
      if ( ret != OK )
        return ret;
      break;
    }

  return OK;
}

static DXLError DXLRemoveValueHandler_NEW( DXLConnection     *conn,
                                           const char        *label )

  /* Used to unregister callable objects in Python as DX value handlers.    */
{
  PyDXLHdlr *hdlr_def;
  int        i;
  DXLError   ret;

  /*printf( "VAL UNREGISTER: %p %s\n", conn, label );*/

  for ( i = 0; i < PyDXLHandler.num; i++ )
    if (( PyDXLHandler.list[i]->conn        == conn         ) &&
        ( PyDXLHandler.list[i]->cb_type     == Hdlr_Value   ) &&
        ( strcmp( PyDXLHandler.list[i]->label, label ) == 0 ))
      break;

  if ( i >= PyDXLHandler.num ) {
    ret = ERROR;
    strcpy( my_error_msg, "No matching callback registered" );
  }
  else {

    ret = DXLRemoveValueHandler( conn, label );
    if ( ret != OK )
      strcpy( my_error_msg, "DXLRemoveValueHandler returned error" );
    else
      ret = PythonDXLDeleteHandler( PyDXLHandler.list[i] );
  }
  return ret;
}


static DXLError DXLSetBrokenConnectionCallback_NEW( DXLConnection     *conn, 
                                                    PyObject          *pyfunc )
  /* Used to register callable objects in Python as DX broken connection     */
  /*   error handlers.  To pass client data, use bound methods, lambdas,     */
  /*   callable objs, etc.                                                   */
{
  PyDXLHdlr *hdlr_def;
  int        i;
  DXLError   ret;

  /*printf( "BCONN REGISTER: %p %p\n", conn, pyfunc );*/

  /*  Initialize record  */
  hdlr_def = PythonDXLNewHandler();
  if ( !hdlr_def )
    return ERROR;

  hdlr_def->cb_type     = Hdlr_BrokenConnection;
  hdlr_def->conn        = conn;
  hdlr_def->py_callback = pyfunc;
  Py_INCREF( hdlr_def->py_callback );

  /*  Register with DX   */
  DXLSetBrokenConnectionCallback( conn, PythonDXLBrokenConnectionHandler, 
                                  hdlr_def );

  /*  Is there a prior matching handler?  If so, nuke it.  */
  for ( i = 0; i < PyDXLHandler.num; i++ )
    if (( PyDXLHandler.list[i]          != hdlr_def   ) &&
        ( PyDXLHandler.list[i]->conn    == conn       ) &&
        ( PyDXLHandler.list[i]->cb_type == Hdlr_BrokenConnection )) {

      ret = PythonDXLDeleteHandler( PyDXLHandler.list[i] );
      if ( ret != OK )
        return ret;
      break;
    }

  return OK;
}

%}
