#!/usr/bin/env python
#
#  6-msg_hdlr.py  -  Test of message handler callbacks.
#

import sys, os, DX
from DX.DXL import *

#----------------------------------------------------------------------------

def DXMsgHandlerCB( conn, msg, packet_type ):
  """  Handle received DX MessageHandler messages  """

  messages = { PACK_INTERRUPT  : "INTERRUPT ",
               PACK_MACRODEF   : "MACRODEF  ",
               PACK_FOREGROUND : "FOREGROUND",
               PACK_BACKGROUND : "BACKGROUND",
               PACK_ERROR      : "ERROR     ",
               PACK_MESSAGE    : "MESSAGE   ",
               PACK_INFO       : "INFO      ",
               PACK_LINQUIRY   : "LINQUIRY  ",
               PACK_LRESPONSE  : "LRESPONSE ",
               PACK_COMPLETE   : "COMPLETE  ",
               PACK_LINK       : "LINK      " }

  print "<DX Msg> %s:  %s" % ( messages[ packet_type ], msg )

#----------------------------------------------------------------------------

def EnableMessageHandler( conn, enable ):
  #  NOTE: On new connections, a default PACK_ERROR handler is plugged in.
  #        But we can unplug it if we want, or plug something else in instead

  def BuildCB( type ):
    def _( conn, msg, type=type ):
      DXMsgHandlerCB( conn, msg, type )
    return _

  for i in range( DXL_PACKETTYPE_LAST+1 ):
    if enable:
      DXLSetMessageHandler   ( conn, i, None, BuildCB( i ) )
    else:
      DXLRemoveMessageHandler( conn, i, None )


#----------------------------------------------------------------------------

def SimpleDXRun():
  
  conn = DXLStartDX( DX.DXEXEC + " -execonly", None )
  DXLSetErrorHandler( conn, lambda conn, msg: None )      # Ignore errors
  EnableMessageHandler  ( conn, 1 )

  DXLSend( conn, "num = 1 + 2;" )
  DXLSend( conn, "DXLOutput( \"result\", num );" )
  DXLExecuteOnce ( conn )
  WaitForDXIdle  ( conn )

  EnableMessageHandler( conn, 0 )
  DXLExitDX      ( conn )

#----------------------------------------------------------------------------

if __name__ == "__main__":

  SimpleDXRun()
